import os
os.environ["KMP_DUPLICATE_LIB_OK"]="TRUE"
from mmseg.apis import inference_segmentor, init_segmentor
from mmseg.datasets.pipelines import Compose
from mmcv.parallel import collate, scatter
from mmseg.models import build_segmentor
from mmcv.runner import load_checkpoint
import mmcv
import cv2
import torch
import numpy as np
import matplotlib.pyplot as plt
import torch.nn.functional as F
import argparse


parser = argparse.ArgumentParser(description='inputs')
parser.add_argument('--net', type=str, default='FCN',
      help='which method to use')
opt = parser.parse_args()



#select the model
model=opt.net
if model=='FCN':
    config_file = 'config/fcn_r101-d8_512x512_160k_ade20k.py'
    checkpoint_file = 'checkpoints/fcn_r101-d8_512x512_160k_ade20k_20200615_105816-fd192bd5.pth'
elif model=='PSPnet':
    config_file = 'config/pspnet_r101-d8_512x512_160k_ade20k.py'
    checkpoint_file = 'checkpoints/pspnet_r101-d8_512x512_160k_ade20k_20200615_100650-967c316f.pth'
elif model=='Deeplabv3plus':
    config_file = 'config/deeplabv3plus_r101-d8_512x512_160k_ade20k.py'
    checkpoint_file = 'checkpoints/deeplabv3plus_r101-d8_512x512_160k_ade20k_20200615_123232-38ed86bb.pth'
elif model=='SETR':
    config_file = 'config/setr_naive_512x512_160k_b16_ade20k.py'
    checkpoint_file = 'checkpoints/setr_naive_512x512_160k_b16_ade20k_20210619_191258-061f24f5.pth'
elif model=='Segformer':
    config_file = 'config/segformer_mit-b4_512x512_160k_ade20k.py'
    checkpoint_file = 'checkpoints/segformer_mit-b4_512x512_160k_ade20k_20210728_183055-7f509d7d.pth'
ROOT = os.path.dirname(os.path.realpath(__file__))
config_file=os.path.join(ROOT, config_file)
checkpoint_file=os.path.join(ROOT, checkpoint_file)

# batch norm
def SyncBN2BN(config):
    for key in config.keys():
        if key== 'norm_cfg':
            if config['norm_cfg']['type']=='SyncBN':
                config['norm_cfg']['type']='BN'
        elif isinstance(config[key],dict):
            config[key]=SyncBN2BN(config[key])
    return config

#open the model
device='cpu'
#device='cuda:0'
# build the model from a config file and a checkpoint file
if device == 'cuda:0':
    model = init_segmentor(config_file, checkpoint_file, device=device)
else :
    if isinstance(config_file, str):
        config = mmcv.Config.fromfile(config_file)
    elif not isinstance(config, mmcv.Config):
        raise TypeError('config must be a filename or Config object, '
                        'but got {}'.format(type(config)))
    config.model.pretrained = None
    config.model.train_cfg = None
    config.model=SyncBN2BN(config.model)
    model = build_segmentor(config.model, test_cfg=config.get('test_cfg'))
    if checkpoint_file is not None:
        checkpoint = load_checkpoint(model, checkpoint_file, map_location='cpu')
        model.CLASSES = checkpoint['meta']['CLASSES']
        model.PALETTE = checkpoint['meta']['PALETTE']
    model.cfg = config  # save the config in the model for convenience
    model.to(device)
    model.eval()


#open and transforme the image in the good format
img='input_0.png'
cfg = model.cfg
Transforme=Compose(cfg.data.test.pipeline[1:])
I=cv2.imread(img)
h,w,c=I.shape
# I=(10 + np.random.randint(0,30,1)[0]*np.random.randn(h,w,c)).clip(0,255).astype(np.uint8)


data={'img':I,'filename':None,'ori_filename':None,'ori_shape':I.shape,'img_shape':I.shape}
data=Transforme(data)
data = collate([data], samples_per_gpu=1)

if next(model.parameters()).is_cuda:
    # scatter to specified GPU
    data = scatter(data, [device])[0]
else :
    data['img_metas'][0]=data['img_metas'][0].data[0]
    
    
#apply the model to the image
with torch.no_grad():
    SoftMax=F.softmax(model.whole_inference(data['img'][0],data['img_metas'][0],rescale='False'),dim=1)

#compute the entropy of the Softmax
SoftMax=(SoftMax.cpu().numpy()[0])
result=[np.argmax(SoftMax,axis=0)]
Detected_labels=np.unique(result[0])
Entropy=np.sum(-np.where(SoftMax>0,SoftMax*np.log(SoftMax),0),axis=0)

state = np.random.get_state()
np.random.seed(42)
# random palette
palette = np.random.randint(
    0, 255, size=(len(model.CLASSES), 3))
np.random.set_state(state)
palette = np.array(palette)
h,w,c=I.shape
color_seg=np.zeros((h,w,3),dtype=int)

label_names=model.CLASSES
text_size_max=0
for label in Detected_labels:
    color_seg[result[0]==label,:]=palette[label]
    text_size,h=cv2.getTextSize(label_names[label], 0, 1, 2)[0]
    if text_size>text_size_max:
        text_size_max=text_size

label_im=255*np.ones(((2*h)*len(Detected_labels), 2*h + text_size_max,3),dtype=np.uint8)
for i,label in enumerate(Detected_labels):
    label_im[2*h*i:2*h*i + 2*h,: 2*h,:]=(palette[label]).astype(np.uint8)
for i,label in enumerate(Detected_labels):
    label_im=cv2.putText(label_im, label_names[label], (2*h + 1,i*2*h +int(1.5*h)), 0, 1, (0,0,0),2)
    
    

cv2.imwrite('img.png',I)
Entropy=(255*Entropy/Entropy.max()).astype(np.uint8)
# Entropy=(255*Entropy/(-len(model.CLASSES)*0.5*np.log(0.5))).astype(np.uint8)
cv2.imwrite('entropy.png',Entropy)
cv2.imwrite('seg.png',color_seg[:,:,::-1]),
cv2.imwrite('label.png',label_im[:,:,::-1])
